from datetime import datetime
from typing import Any, Dict, Generator, List, Optional, Union

from gitlab.base import RESTObject
from gitlab.v4.objects import Project, ProjectBranch, ProjectCommit
from gitlab.v4.objects import User as GitlabUser

from jf_ingest.config import GitConfig
from jf_ingest.jf_git.adapters import GitAdapter
from jf_ingest.jf_git.clients.gitlab import GitlabClient
from jf_ingest.jf_git.standardized_models import (
    StandardizedBranch,
    StandardizedCommit,
    StandardizedOrganization,
    StandardizedPullRequest,
    StandardizedPullRequestMetadata,
    StandardizedRepository,
    StandardizedTeam,
    StandardizedUser,
)
from jf_ingest.utils import parse_gitlab_date


class GitlabAdapter(GitAdapter):

    def __init__(self, config: GitConfig):
        self.config = config
        self.client = GitlabClient(auth_config=config.git_auth_config)
        self.group_id_to_full_path: Dict[str, str] = {}

    def get_api_scopes(self) -> str:
        """Return the list of API Scopes. This is useful for Validation

        Returns:
            str: A string of API scopes we have, given the adapters credentials
        """
        raise NotImplementedError()

    @staticmethod
    def _get_group_id_from_gid(gitlab_gid: str) -> str:
        """Helper function.
        The Gitlab GQL returns Group (Organization) IDs with this weird GID format.
        All we care about is the number trailing at the end.
        Gitlab Format: gid://gitlab/Group/{ID_NUMBER}
        """
        return gitlab_gid.split('gid://gitlab/Group/')[1]

    @staticmethod
    def _get_project_id_from_gid(gitlab_gid: str) -> str:
        """Helper function.
        The Gitlab GQL returns Project (Organization) IDs with this weird GID format.
        All we care about is the number trailing at the end.
        Gitlab Format: gid://gitlab/Project/{ID_NUMBER}
        """
        return gitlab_gid.split('gid://gitlab/Project/')[1]

    @staticmethod
    def _standardize_commit_author(api_user: dict) -> StandardizedUser:
        id = api_user.get('id') or api_user.get('email')  # Commit users may not have ids
        name = api_user.get('name')
        login = api_user.get('login') or api_user.get('email')
        email = api_user.get('email')

        return StandardizedUser(
            id=str(id),
            login=str(login),
            name=str(name),
            email=str(email),
        )

    @staticmethod
    def _standardize_gitlab_commit(
        gitlab_commit: ProjectCommit | RESTObject,
        standardized_repo: StandardizedRepository,
        branch_name: str,
        strip_text_content: bool,
        redact_names_and_urls: bool,
    ) -> StandardizedCommit:
        """
        Converts a ProjectCommit from the gitlab client to a standardized jf commit.
        """
        commit_url = gitlab_commit.web_url if not redact_names_and_urls else None
        author = GitlabAdapter._standardize_commit_author(
            {
                'name': gitlab_commit.author_name,
                'email': gitlab_commit.author_email,
            }
        )
        return StandardizedCommit(
            hash=gitlab_commit.id,
            author=author,
            url=commit_url,
            commit_date=parse_gitlab_date(gitlab_commit.committed_date),
            author_date=parse_gitlab_date(gitlab_commit.authored_date),
            message=GitAdapter.sanitize_text(gitlab_commit.message, strip_text_content),
            is_merge=len(gitlab_commit.parent_ids) > 1,
            repo=standardized_repo.short(),  # use short form of repo
            branch_name=(
                branch_name
                if not redact_names_and_urls
                else GitAdapter.branch_redactor.redact_name(branch_name)
            ),
        )

    @staticmethod
    def _standardize_gitlab_project(
        gitlab_repo: Project | RESTObject, standardized_organization: StandardizedOrganization
    ) -> StandardizedRepository:
        """
        Converts a Project from the gitlab client to a standardized jf project.
        """
        return StandardizedRepository(
            id=str(gitlab_repo.id),
            name=gitlab_repo.name,
            full_name=gitlab_repo.name,
            url=gitlab_repo.web_url,
            default_branch_sha='',
            default_branch_name=gitlab_repo.default_branch,
            organization=standardized_organization,
            is_fork=getattr(gitlab_repo, 'forked_from_project', False),
        )

    @staticmethod
    def _standardize_gitlab_branch(
        gitlab_branch: ProjectBranch | RESTObject, standardized_repository: StandardizedRepository
    ) -> StandardizedBranch:
        """
        Converts a Branch (ProjectBranch) from the gitlab client to a standardized jf branch.
        """
        return StandardizedBranch(
            name=gitlab_branch.name,
            sha=gitlab_branch.commit['id'],
            repo_id=standardized_repository.id,
            is_default=gitlab_branch.default,
        )

    @staticmethod
    def _standardize_user(api_user: Union[RESTObject, GitlabUser]) -> StandardizedUser:
        return StandardizedUser(
            id=api_user.id,
            name=api_user.name,
            login=api_user.username,
            email=None,
            url=api_user.web_url,
        )

    def get_group_full_path_from_id(self, group_id: str) -> str:
        if group_id not in self.group_id_to_full_path:
            _, full_path, _ = self.client.get_organization_name_full_path_and_url(login=group_id)
            self.group_id_to_full_path[group_id] = full_path

        return self.group_id_to_full_path[group_id]

    def get_group_full_path_from_organization(self, org: StandardizedOrganization) -> str:
        return self.get_group_full_path_from_id(org.login)

    def get_organizations(self) -> List[StandardizedOrganization]:
        """Get the list of organizations the adapter has access to

        Returns:
            List[StandardizedOrganization]: A list of standardized organizations within this Git Instance
        """
        orgs: List[StandardizedOrganization] = []
        if not self.config.discover_organizations:
            for group_id in self.config.git_organizations:
                name, full_path, url = self.client.get_organization_name_full_path_and_url(
                    login=group_id
                )
                self.group_id_to_full_path[group_id] = full_path
                orgs.append(
                    StandardizedOrganization(id=group_id, name=name, login=group_id, url=url)
                )
        else:
            # Discover Orgs
            for api_org in self.client.get_organizations_gql():
                group_id = self._get_group_id_from_gid(api_org['groupIdStr'])
                full_path = api_org['fullPath']
                self.group_id_to_full_path[group_id] = full_path
                orgs.append(
                    StandardizedOrganization(
                        id=group_id, name=api_org['name'], login=group_id, url=api_org['webUrl']
                    )
                )

        return orgs

    def get_users(
        self, standardized_organization: StandardizedOrganization, limit: Optional[int] = None
    ) -> Generator[StandardizedUser, None, None]:
        """
        Get all users in a given Git Organization

        Args:
            standardized_organization (StandardizedOrganization): A standardized Git Organization Object
            limit (int, optional): When provided, only returns this many users. Defaults to None.

        Returns:
            Generator[StandardizedUser, None, None]: A generator of StandardizedUser objects
        """
        for idx, user in enumerate(self.client.get_users(group_id=standardized_organization.login)):
            if limit and idx >= limit:
                return

            yield self._standardize_user(user)

    def get_teams(
        self, standardized_organization: StandardizedOrganization, limit: Optional[int] = None
    ) -> Generator[StandardizedTeam, None, None]:
        """
        This function is to align with what the parent adapter class expects.
        GitLab does not have a concept of teams past groups, which we use as organizations.
        This will return an empty list, regardless of arguments.
        """
        teams: List[StandardizedTeam] = []
        yield from teams

    def get_repos(
        self, standardized_organization: StandardizedOrganization, use_gql: Optional[bool] = False
    ) -> Generator[StandardizedRepository, None, None]:
        """Get a list of standardized repositories within a given organization

        Args:
            standardized_organization (StandardizedOrganization): A standardized organization
            use_gql (bool, optional): Whether to use GQL queries. Defaults to False.

        Returns:
            Generator[StandardizedRepository]: An iterable of standardized Repositories
        """
        if use_gql:
            full_path = self.get_group_full_path_from_organization(standardized_organization)
            for repo in self.client.get_repos_gql(full_path):
                yield StandardizedRepository(
                    id=self._get_project_id_from_gid(repo['id']),
                    name=repo['name'],
                    full_name=repo['name'],
                    url=repo['webUrl'],
                    default_branch_sha='',
                    default_branch_name=repo['repository']['defaultBranchName'],
                    organization=standardized_organization,
                    is_fork=repo['isForked'],
                )
        else:
            for client_repo in self.client.get_repos(standardized_organization):
                yield self._standardize_gitlab_project(client_repo, standardized_organization)

    def get_commits_for_default_branch(
        self,
        standardized_repo: StandardizedRepository,
        limit: Optional[int] = None,
        pull_since: Optional[datetime] = None,
        pull_until: Optional[datetime] = None,
    ) -> Generator[StandardizedCommit, None, None]:
        """For a given repo, get all the commits that are on the Default Branch.

        Args:
            standardized_repo (StandardizedRepository): A standard Repository object
            limit (int): limit the number of commit objects we will yield
            pull_since (datetime): filter commits to be newer than this date
            pull_until (datetime): filter commits to be older than this date

        Returns:
            Generator[StandardizedCommit]: An iterable of standardized commits
        """
        default_branch_name: Optional[str] = standardized_repo.default_branch_name
        if default_branch_name:
            for j, api_commit in enumerate(
                self.client.get_commits(
                    standardized_repo,
                    since=pull_since,
                    branch_name=default_branch_name,
                ),
                start=1,
            ):
                yield self._standardize_gitlab_commit(
                    api_commit,
                    standardized_repo,
                    default_branch_name,
                    self.config.git_strip_text_content,
                    self.config.git_redact_names_and_urls,
                )
                if limit and j >= limit:
                    return

    def get_branches_for_repo(
        self,
        standardized_repo: StandardizedRepository,
        pull_branches: Optional[bool] = False,
    ) -> Generator[StandardizedBranch, None, None]:
        """Function for pulling branches for a repository. By default, pull_branches will run as False,
        so we will only process the default branch. If pull_branches is true, than we will pull all
        branches in this repository

        Args:
            standardized_repo (StandardizedRepository): A standardized repo, which hold info about the default branch.
            pull_branches (bool): A boolean flag. If True, pull all branches available on Repo. If false, only process the default branch. Defaults to False.

        Yields:
            StandardizedBranch: A Standardized Branch Object
        """
        search_term: Optional[str] = None
        if not pull_branches:
            search_term = f'^{standardized_repo.default_branch_name}$'

        for api_branch in self.client.get_branches_for_repo(
            standardized_repo, search_term=search_term
        ):
            yield self._standardize_gitlab_branch(api_branch, standardized_repo)

    def get_commits_for_branches(
        self,
        standardized_repo: StandardizedRepository,
        branches: List[StandardizedBranch],
        pull_since: Optional[datetime] = None,
        pull_until: Optional[datetime] = None,
    ) -> Generator[StandardizedCommit, None, None]:
        """For a given repo, get all the commits that are on the included branches.
        Included branches are found by crawling across the branches pulled/available
        from get_filtered_branches

        Args:
            standardized_repo (StandardizedRepository): A standard Repository object
            pull_since (datetime): A date to pull from
            pull_until (datetime): A date to pull up to

        Returns:
            List[StandardizedCommit]: A list of standardized commits
        """
        raise NotImplementedError()

    def get_pr_metadata(
        self,
        standardized_repo: StandardizedRepository,
        limit: Optional[int] = None,
        pr_pull_from_date: Optional[datetime] = None,
    ) -> Generator[StandardizedPullRequestMetadata, None, None]:
        """Get all PRs, but only included the bare necesaties

        Args:
            standardized_repo (StandardizedRepository): A standardized repository
            limit (int, optional): When provided, the number of items returned is limited.
                Useful for the validation use case, where we want to just verify we can pull PRs.
                Defaults to None.
            pr_pull_from_date: This is currently only used by the GithubAdapter. Probably won't be useful for this adapter

        Returns:
            List[StandardizedPullRequest]: A list of standardized PRs
        """
        raise NotImplementedError()

    def git_provider_pr_endpoint_supports_date_filtering(self) -> bool:
        """Returns a boolean on if this PR supports time window filtering.
        So far, Github DOES NOT support this (it's adapter will return False)
        but ADO does support this (it's adapter will return True)

        Returns:
            bool: A boolean on if the adapter supports time filtering when searching for PRs
        """
        return True

    def get_prs(
        self,
        standardized_repo: StandardizedRepository,
        pull_files_for_pr: bool = False,
        hash_files_for_prs: bool = False,
        limit: Optional[int] = None,
        start_cursor: Optional[Any] = None,
        start_window: Optional[datetime] = None,
        end_window: Optional[datetime] = None,
    ) -> Generator[StandardizedPullRequest, None, None]:
        """Get the list of standardized Pull Requests for a Standardized Repository.

        Args:
            standardized_repo (StandardizedRepository): A standardized repository
            pull_files_for_pr (bool): When provided, we will pull file metadata for all PRs
            hash_files_for_prs (bool): When provided, all file metadata will be hashed for PRs
            limit (int, optional): When provided, the number of items returned is limited.
                Useful for the validation use case, where we want to just verify we can pull PRs.
                Defaults to None.

        Returns:
            List[StandardizedPullRequest]: A list of standardized PRs
        """
        raise NotImplementedError()



import json
import logging
from datetime import datetime
from typing import Any, Dict, Generator, Optional, Tuple, Union

import requests
from gitlab import Gitlab
from gitlab.base import RESTObject
from gitlab.v4.objects import Project, ProjectBranch, ProjectCommit
from gitlab.v4.objects import User as GitlabUser

from jf_ingest.config import GitAuthConfig
from jf_ingest.constants import Constants
from jf_ingest.graphql_utils import GQL_PAGE_INFO_BLOCK
from jf_ingest.jf_git.standardized_models import (
    StandardizedOrganization,
    StandardizedRepository,
)
from jf_ingest.utils import retry_for_status, retry_session

logger = logging.getLogger(__name__)


class GitlabClient:
    # NOTE: We currently have some functions in this class that access GraphQL.
    # When trying to implement that, we ran into some authentication issues and limitations that moved us toward
    #   using the rest client instead.
    # Functions that make use of GraphQL are suffixed with _gql.
    # If we want to keep them, we'd ideally eventually move them to their own subclass of the client.
    GITLAB_GQL_USER_FRAGMENT = "... on User {login, id: databaseId, email, name, url}"

    def __init__(self, auth_config: GitAuthConfig, **kwargs):
        """Gitlab Client, used as a wrapper for getting raw data from the API.
        This client will get data mainly from the GraphQL API endpoints, although
        it will use the REST API endpoints for a small amount of functions as well.

        Args:
            auth_config (GitAuthConfig): A valid GitAuthConfiguration object
            kwargs: kwargs are used to pass arguments to the inner Session object, if no session object is provided as part of the GitAuthConfig
        """
        self.company_slug: str = auth_config.company_slug
        self.rest_api_url: Optional[str] = auth_config.base_url
        self.gql_base_url: str = f'{auth_config.base_url}/api/graphql'
        if session := auth_config.session:
            self.session: requests.Session = session
        else:
            self.session = retry_session(**kwargs)
            self.session.headers.update(
                {
                    'Authorization': f'Bearer {auth_config.token}',
                    'Content-Type': 'application/json',
                    'User-Agent': f'{Constants.JELLYFISH_USER_AGENT} ({requests.utils.default_user_agent()})',
                }
            )
        self.session.verify = auth_config.verify
        self.client: Gitlab = Gitlab(url=self.rest_api_url, session=self.session)

    def get_organization_name_full_path_and_url(self, login: str) -> Tuple[str, str, str]:
        """In Jellyfish Land, the JFGithubOrganization is the normalization of Github Organizations,
        AzureDevops Organizations, Bitbucket Projects, and Gitlab Groups. The login field is the unique
        key. For Gitlab Groups, we set the login to the be the Group ID, which is a numeric value.
        The GraphQL Group Queries accept a "fullPath" argument, and NOT the Group ID. If we only have
        the GroupID (set by the login value), then this helper function can be used to translate the
        GroupID to a Full Path.
        NOTE: For performance reasons, we should probably graph the FullPath when we query GraphQL for
        Groups in general, and then cache those values. We should NOT call this function everytime,
        because it could have performance implications

        Args:
            login (str): The JFGithubOrganization login, which is the Group ID in Gitlab land
        Returns:
            name, full_path, url (str, str, str): The name, Full Path, and url for this gitlab Group
        """
        group_url = f'{self.rest_api_url}/api/v4/groups/{login}?with_projects=False'
        response: requests.Response = retry_for_status(self.session.get, url=group_url)
        response.raise_for_status()
        response_json = response.json()
        return (
            str(response_json['name']),
            str(response_json['full_path']),
            str(response_json['web_url']),
        )

    def get_raw_result_gql(self, query_body: str, max_attempts: int = 7) -> Dict:
        """Gets the raw results from a Graphql Query.

        Args:
            query_body (str): A query body to hit GQL with
            max_attempts (int, optional): The number of retries we should make when we specifically run into GQL Rate limiting. This value is important if the GQL endpoint doesn't give us (or gives us a malformed) rate limit header. Defaults to 7.

        Raises:
            GqlRateLimitExceededException: A custom exception if we run into GQL rate limiting and we run out of attempts (based on max_attempts)
            Exception: Any other random exception we encounter, although the big rate limiting use cases are generally covered

        Returns:
            dict: A raw dictionary result from GQL
        """
        response: requests.Response = retry_for_status(
            self.session.post,
            url=self.gql_base_url,
            json={'query': query_body},
            max_retries_for_retry_for_status=max_attempts,
        )
        json_str = response.content.decode()
        json_data: Dict = json.loads(json_str)
        if error_list_dict := json_data.get('errors'):
            error_message = ','.join([error_dict.get('message') for error_dict in error_list_dict])
            raise Exception(f'An Error occurred when attempting to query GraphQL: {error_message}')

        return json_data

    def page_results_gql(
        self, query_body: str, path_to_page_info: str, cursor: str = 'null'
    ) -> Generator[Dict, None, None]:
        """This is a helper function for paging results from GraphQL. It expects
        a query body to hit Graphql with that has a %s marker after the "after:"
        key word, so that we can inject a cursor into the query. This will allow
        us to page results in GraphQL.
        To use this function properly, the section you are trying to page MUST
        INCLUDE VALID PAGE INFO (including the hasNext and endCursor attributes)

        Args:
            query_body (str): The query body to hit GraphQL with
            path_to_page_info (str): A string of period separated words that lead
            to the part of the query that we are trying to page. Example: data.organization.userQuery
            cursor (str, optional): LEAVE AS NULL - this argument is use recursively to page. The cursor
            will continuously go up, based on the endCursor attribute in the GQL call. Defaults to 'null'.

        Yields:
            Generator[dict, None, None]: This function yields each item from all the pages paged, item by item
        """
        hasNextPage = True
        while hasNextPage:
            # Fetch results
            result = self.get_raw_result_gql(query_body=(query_body % cursor))

            yield result

            # Get relevant data and yield it
            path_tokens = path_to_page_info.split('.')
            for token in path_tokens:
                result = result[token]

            page_info = result['pageInfo']
            # Need to grab the cursor and wrap it in quotes
            _cursor = page_info['endCursor']
            # If endCursor returns null (None), break out of loop
            hasNextPage = page_info['hasNextPage'] and _cursor
            cursor = f'"{_cursor}"'

    def get_teams(self, *args, **kwargs) -> list:
        """
        This function is to align with other clients.
        GitLab does not have a concept of teams past groups, which we use as organizations.
        This will return an empty list, regardless of arguments.
        """
        return []

    def get_repos(
        self, jf_org: StandardizedOrganization
    ) -> Generator[Project | RESTObject, None, None]:
        group = self.client.groups.get(jf_org.login)

        for repo in group.projects.list(iterator=True, include_subgroups=True):
            yield repo

    def get_commits(
        self,
        jf_repo: StandardizedRepository,
        since: Optional[datetime] = None,
        until: Optional[datetime] = None,
        branch_name: Optional[str] = None,
    ) -> Generator[ProjectCommit | RESTObject, None, None]:
        project_id = jf_repo.id
        project = self.client.projects.get(project_id)

        kwargs_dict: dict[str, Any] = {
            'iterator': True,  # Returns a generator
        }
        if branch_name:
            kwargs_dict['ref_name'] = branch_name
        if since:
            kwargs_dict['since'] = since
        if until:
            kwargs_dict['until'] = until

        for commit in project.commits.list(**kwargs_dict):
            yield commit

    def get_branches_for_repo(
        self, jf_repo: StandardizedRepository, search_term: Optional[str] = None
    ) -> Generator[ProjectBranch | RESTObject, None, None]:
        project_id = jf_repo.id
        project = self.client.projects.get(project_id)

        # The GitLab client might take `search=None` as something to be passed to the request, only add that key
        # if we have a search term.
        kwargs_dict: dict[str, Any] = {'iterator': True}
        if search_term:
            kwargs_dict['search'] = search_term

        for branch in project.branches.list(**kwargs_dict):
            yield branch

    def get_organizations_gql(
        self, page_size: int = 100, sort_key: str = 'id_asc'
    ) -> Generator[Dict, None, None]:
        query_body = f"""
        {{
            groupsQuery: groups(first: {page_size}, sort: "{sort_key}", after: %s){{
                {GQL_PAGE_INFO_BLOCK}
                groups: nodes {{
                    groupIdStr: id
                    name
                    fullPath
                    webUrl
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.groupsQuery'
        ):
            for group in page['data']['groupsQuery']['groups']:
                yield group

    def get_repos_gql(
        self, group_full_path: str, page_size: int = 100
    ) -> Generator[Dict, None, None]:
        query_body = f"""
        {{
            group(fullPath: "{group_full_path}") {{
                projectsQuery: projects(first: {page_size}, after: %s) {{
                    {GQL_PAGE_INFO_BLOCK}
                    projects: nodes {{
                        ... on Project {{
                            name,
                            webUrl,
                            description,
                            isForked,
                            repository {{
                                ... on Repository {{
                                    defaultBranchName: rootRef
                                }}
                            }}
                        }}
                    }}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.group.projectsQuery'
        ):
            for project in page['data']['group']['projectsQuery']['projects']:
                yield project

    def get_users(self, group_id: str) -> Generator[Union[RESTObject, GitlabUser], None, None]:
        """
        Gets all users for a given Gitlab group (aka organization)

        Args:
            group_id: ID of the group (organization) to get users for

        Returns:
            Generator[Union[RESTObject, GitlabUser], None, None]: Generator yielding one user object at a time
        """
        group = self.client.groups.get(group_id)

        for user in group.members.list(all=True, iterator=True):
            yield user




import json
from jf_ingest.jf_jira.utils import JiraFieldIdentifier, expand_and_normalize_jira_fields
from tests.jf_jira.utils import get_fixture_file_data

def get_field_data() -> list[dict]:
    return [
    {
        "id": "statuscategorychangedate",
        "key": "statuscategorychangedate",
        "name": "Status Category Changed",
        "custom": False,
        "orderable": False,
        "navigable": True,
        "searchable": True,
        "clauseNames": [
            "statusCategoryChangedDate"
        ],
        "schema": {
            "type": "datetime",
            "system": "statuscategorychangedate"
        }
    },
    {
        "id": "parent",
        "key": "parent",
        "name": "Parent",
        "custom": False,
        "orderable": False,
        "navigable": True,
        "searchable": False,
        "clauseNames": [
            "parent"
        ]
    },
    {
        "id": "customfield_10070",
        "key": "customfield_10070",
        "name": "Satisfaction date",
        "untranslatedName": "Satisfaction date",
        "custom": True,
        "orderable": True,
        "navigable": True,
        "searchable": True,
        "clauseNames": [
            "cf[10070]",
            "Satisfaction date"
        ],
        "schema": {
            "type": "datetime",
            "custom": "com.atlassian.servicedesk:sd-request-feedback-date",
            "customId": 10070
        }
    },
    {
        "id": "customfield_10071",
        "key": "customfield_10071",
        "name": "Reporters",
        "untranslatedName": "Reporters",
        "custom": True,
        "orderable": True,
        "navigable": True,
        "searchable": True,
        "clauseNames": [
            "Reporters",
            "Reporters[User Picker (multiple users)]",
            "cf[10071]"
        ],
        "schema": {
            "type": "array",
            "items": "user",
            "custom": "com.atlassian.jira.plugin.system.customfieldtypes:multiuserpicker",
            "customId": 10071
        }
    }
]

def test_expand_and_normalize_jira_fields_smoke_test():
    raw_fields = get_field_data()
    
    jira_field_identifiers = expand_and_normalize_jira_fields(fields_from_jira=raw_fields, field_names_or_ids=[])
    assert jira_field_identifiers == []
    

def test_expand_and_normalize_jira_fields_with_id():
    raw_fields = get_field_data()
    
    fields_to_expand = ['customfield_10070']
    jira_field_identifiers = expand_and_normalize_jira_fields(fields_from_jira=raw_fields, field_names_or_ids=fields_to_expand)
    assert len(jira_field_identifiers) == 1
    assert jira_field_identifiers[0].jira_field_id == 'customfield_10070'
    assert jira_field_identifiers[0].jira_field_name == 'Satisfaction date'

def test_expand_and_normalize_jira_fields_with_name():
    raw_fields = get_field_data()
    
    fields_to_expand = ['Satisfaction date']
    jira_field_identifiers = expand_and_normalize_jira_fields(fields_from_jira=raw_fields, field_names_or_ids=fields_to_expand)
    assert len(jira_field_identifiers) == 1
    assert jira_field_identifiers[0].jira_field_id == 'customfield_10070'
    assert jira_field_identifiers[0].jira_field_name == 'Satisfaction date'

def test_expand_and_normalize_jira_fields_with_redundant_data():
    raw_fields = get_field_data()
    
    fields_to_expand = ['customfield_10070', 'Satisfaction date']
    jira_field_identifiers = expand_and_normalize_jira_fields(fields_from_jira=raw_fields, field_names_or_ids=fields_to_expand)
    assert len(jira_field_identifiers) == 1
    assert jira_field_identifiers[0].jira_field_id == 'customfield_10070'
    assert jira_field_identifiers[0].jira_field_name == 'Satisfaction date'
    
def test_expand_and_normalize_jira_fields_with_all_ids():
    raw_fields = sorted(get_field_data(), key=lambda _dict: _dict['id'])
    ids = [f['id'] for f in raw_fields]
    names = [f['name'] for f in raw_fields]
    assert len(ids) == len(names)
    
    
    jira_field_identifiers = expand_and_normalize_jira_fields(fields_from_jira=raw_fields, field_names_or_ids=ids)
    
    assert len(jira_field_identifiers) == len(ids)
    for id, name, jira_field_identifier in zip(ids, names, sorted(jira_field_identifiers, key=lambda jfi: jfi.jira_field_id)):
        print(id, name, jira_field_identifier)
        assert jira_field_identifier.jira_field_id == id
        assert jira_field_identifier.jira_field_name == name

def test_expand_and_normalize_jira_fields_with_all_names():
    raw_fields = sorted(get_field_data(), key=lambda _dict: _dict['id'])
    ids = [f['id'] for f in raw_fields]
    names = [f['name'] for f in raw_fields]
    assert len(ids) == len(names)
    
    
    jira_field_identifiers = expand_and_normalize_jira_fields(fields_from_jira=raw_fields, field_names_or_ids=names)
    
    assert len(jira_field_identifiers) == len(names)
    for id, name, jira_field_identifier in zip(ids, names, sorted(jira_field_identifiers, key=lambda jfi: jfi.jira_field_id)):
        print(id, name, jira_field_identifier)
        assert jira_field_identifier.jira_field_id == id
        assert jira_field_identifier.jira_field_name == name
        
def test_matches_changelog_item_base():
    field_name = 'FieldName'
    field_id = 'customfield_xxxxx'
    changelog_item = {
        'field': field_name,
        'fieldId': field_id
    }
    
    jfi = JiraFieldIdentifier(jira_field_id=field_id, jira_field_name=field_name)
    assert jfi.matches_changelog_item(changelog_item)
    
    changelog_item_no_id = {
        'field': field_name
    }
    
    assert jfi.matches_changelog_item(changelog_item_no_id)
    
    changelog_item_no_name = {
        'fieldId': field_id
    }
    assert jfi.matches_changelog_item(changelog_item_no_name)
    
    changelog_item_reversed = {
        'fieldId': field_name,
        'field': field_id,
    }
    assert not jfi.matches_changelog_item(changelog_item_reversed)
    
    changelog_item_reversed = {
        'fieldId': 'nomatch',
        'field': 'nomatch',
    }
    assert not jfi.matches_changelog_item(changelog_item_reversed)


import logging
import traceback
from datetime import datetime
from typing import Any, Dict, Generator, List, NamedTuple, Optional

from dateutil import parser
from requests.exceptions import HTTPError

from jf_ingest import diagnostics, logging_helper
from jf_ingest.config import GitConfig
from jf_ingest.graphql_utils import gql_format_to_datetime
from jf_ingest.jf_git.adapters import GitAdapter
from jf_ingest.jf_git.clients.github import GithubClient
from jf_ingest.jf_git.standardized_models import (
    PullRequestReviewState,
    StandardizedBranch,
    StandardizedCommit,
    StandardizedFileData,
    StandardizedLabel,
    StandardizedOrganization,
    StandardizedPullRequest,
    StandardizedPullRequestComment,
    StandardizedPullRequestMetadata,
    StandardizedPullRequestReview,
    StandardizedRepository,
    StandardizedShortRepository,
    StandardizedTeam,
    StandardizedUser,
)

logger = logging.getLogger(__name__)

'''

    Data Fetching

'''


class GithubRepoMetaData(NamedTuple):
    """This is a data class unique to Github. It helps us cache information about repos
    so we can intelligently determine if we need to pull commits, branches, and PRs.
    """

    # Repo ID, as set by github
    repo_id: str
    # Branches count
    branches_count: int
    # Commits information
    #  Commits are pulled in Deltas, so we'll have a datetime that we're pulling from.
    #  When initially querying repo data, get the latest commit date. That way, in
    #  memory, we can determine if we need to pull commits or not. For customers with
    #  large repo counts, this can save us A LOT of time!
    latest_commit_date_on_default_branch: Optional[
        datetime
    ]  # If this value is none, that indicates there are NO commits to pull
    # PR Information
    #  PRs follow the same pattern as commits above
    latest_pr_updated_at: Optional[
        datetime
    ]  # If this value is none, that indicates there are NO PRs to pull


class GithubAdapter(GitAdapter):
    def __init__(self, config: GitConfig):
        # Git Config options
        self.client = GithubClient(config.git_auth_config)
        self.config = config
        self.repo_id_to_name_lookup: Dict = {}
        self.repo_id_to_repo_meta_data: Dict[str, GithubRepoMetaData] = {}

    def get_api_scopes(self):
        return self.client.get_scopes_of_api_token()

    @diagnostics.capture_timing()
    @logging_helper.log_entry_exit(logger)
    def get_organizations(self) -> List[StandardizedOrganization]:
        # NOTE: For github, organization equates to a Github organization here!
        # We have the name of the organization in the config, expected to be a list of length one
        try:
            raw_orgs = []
            if self.config.discover_organizations:
                for raw_org in self.client.get_all_organizations():
                    raw_orgs.append(raw_org)
            else:
                for org_login in self.config.git_organizations:
                    raw_orgs.append(self.client.get_organization_by_login(org_login))
        except HTTPError as e:
            if e.response.status_code == 404:
                # give something a little nicer for 404s
                raise ValueError(
                    'Organization not found. Make sure your token has appropriate access to Github.'
                )
            raise
        return [
            _standardize_organization(raw_org, self.config.git_redact_names_and_urls)
            for raw_org in raw_orgs
        ]

    def get_users(
        self, standardized_organization: StandardizedOrganization, limit: Optional[int] = None
    ) -> Generator[StandardizedUser, None, None]:
        for i, user in enumerate(self.client.get_users(standardized_organization.login), start=1):
            if user:
                yield _standardize_user(user)
            if limit and i >= limit:
                return

    def get_teams(
        self, standardized_organization: StandardizedOrganization, limit: Optional[int] = None
    ) -> Generator[StandardizedTeam, None, None]:
        for i, team in enumerate(
            self.client.get_teams(login=standardized_organization.login), start=1
        ):
            yield _standardize_team(team)
            if limit and i >= limit:
                return

    def get_repos(
        self, standardized_organization: StandardizedOrganization
    ) -> Generator[StandardizedRepository, None, None]:
        filters = []
        if self.config.included_repos:
            filters.append(
                lambda r: r['name'].lower() in set([r.lower() for r in self.config.included_repos])
            )
        if self.config.excluded_repos:
            filters.append(
                lambda r: r['name'].lower()
                not in set([r.lower() for r in self.config.excluded_repos])
            )

        for api_repo in self.client.get_repos(
            standardized_organization.login, repo_filters=filters
        ):
            # Enter the repo to the ID to Name look up, incase the repo name gets
            # scrubbed by our git_redact_names_and_urls logic
            repo_id = str(api_repo['id'])
            self.repo_id_to_name_lookup[repo_id] = api_repo['name']

            # Extract info for metadata caching
            # Get Branch Count
            total_branches = int(api_repo['branches']['totalCount'])

            # Get latest commit date
            # NOTE: Default branch can be None!
            latest_commit_date_on_default_branch: Optional[datetime] = None
            if default_branch := api_repo.get('defaultBranch'):
                if commits := default_branch['commitsQuery']['commitHistory']['commits']:
                    latest_commit_date_on_default_branch = gql_format_to_datetime(
                        commits[0]['committedDate']
                    )

            # Get PR Last updated at
            pr_updated_at: Optional[datetime] = None
            pr_query = api_repo['prQuery']
            if prs := pr_query['prs']:
                pr_updated_at = gql_format_to_datetime(prs[0]['updatedAt'])

            self.repo_id_to_repo_meta_data[repo_id] = GithubRepoMetaData(
                repo_id=repo_id,
                branches_count=total_branches,
                latest_commit_date_on_default_branch=latest_commit_date_on_default_branch,
                latest_pr_updated_at=pr_updated_at,
            )

            yield _standardize_repo(
                api_repo, standardized_organization, self.config.git_redact_names_and_urls
            )

    def get_branches_for_repo(
        self,
        standardized_repo: StandardizedRepository,
        pull_branches: Optional[bool] = False,
    ) -> Generator[StandardizedBranch, None, None]:
        # It is possible to have 0 branches on a repo. If that is the case
        # and we've already detected that via our repo metadata query, then
        # we don't need to hit the API for any branch data
        if self.repo_id_to_repo_meta_data[standardized_repo.id].branches_count == 0:
            return
        if pull_branches:
            repo_name = self.repo_id_to_name_lookup[standardized_repo.id]
            for branch in self.client.get_branches(
                login=standardized_repo.organization.login, repo_name=repo_name
            ):
                if standardized_branch := _standardize_branch(
                    branch,
                    repo_id=standardized_repo.id,
                    redact_names_and_urls=self.config.git_redact_names_and_urls,
                    is_default_branch=standardized_repo.default_branch_sha
                    == branch['target']['sha'],
                ):
                    yield standardized_branch
        else:
            # Above, if we're pulling all branches, it's safe to assume that the default branch
            # will be included in that
            # When we don't pull all branches, always return default branch
            if standardized_repo.default_branch_name:
                yield StandardizedBranch(
                    repo_id=standardized_repo.id,
                    name=standardized_repo.default_branch_name,
                    sha=standardized_repo.default_branch_sha,
                    is_default=True,
                )

    def get_commits_for_branches(
        self,
        standardized_repo: StandardizedRepository,
        branches: List[StandardizedBranch],
        pull_since: Optional[datetime] = None,
        pull_until: Optional[datetime] = None,
    ) -> Generator[StandardizedCommit, None, None]:
        if not pull_since:
            logging_helper.send_to_agent_log_file(
                f'When pulling commits for Branches, the "pull_since" argument is required for Github',
                level=logging.ERROR,
            )
            return

        repo_meta_data = self.repo_id_to_repo_meta_data[standardized_repo.id]

        for branch_name in self.get_filtered_branches(standardized_repo, branches):
            try:
                login = standardized_repo.organization.login
                repo_name = self.repo_id_to_name_lookup[standardized_repo.id]

                # Check out repo meta data cache to see
                if standardized_repo.default_branch_name == branch_name and repo_meta_data:
                    # If we've cached that this commit is older or equal to our pull_since value,
                    # do not bother hitting the API for more commit data OR if there are no commits
                    # If repo_meta_data.latest_commit_date_on_default_branch is None, that indicates no commits
                    if not repo_meta_data.latest_commit_date_on_default_branch:
                        logging_helper.send_to_agent_log_file(
                            f'Not pulling commits for branch {branch_name} because this branch has no commits, '
                            'according to the repo meta data.'
                        )
                        continue
                    if repo_meta_data.latest_commit_date_on_default_branch <= pull_since:
                        logging_helper.send_to_agent_log_file(
                            f'Not pulling commits for branch {branch_name} because there are no commits newer than the pull_since data. '
                            f'pull_since: {pull_since}, latest_commit_for_branch: {repo_meta_data.latest_commit_date_on_default_branch}'
                        )
                        continue

                for api_commit in self.client.get_commits(
                    login=login,
                    repo_name=repo_name,
                    branch_name=branch_name,
                    since=pull_since,
                    until=pull_until,
                ):
                    yield _standardize_commit(
                        api_commit,
                        standardized_repo,
                        branch_name,
                        self.config.git_strip_text_content,
                        self.config.git_redact_names_and_urls,
                    )

            except Exception as e:
                logging_helper.send_to_agent_log_file(traceback.format_exc(), level=logging.ERROR)
                logger.warning(f':WARN: Got exception for branch {branch_name}: {e}. Skipping...')

    def get_commits_for_default_branch(
        self,
        standardized_repo: StandardizedRepository,
        limit: Optional[int] = None,
        pull_since: Optional[datetime] = None,
        pull_until: Optional[datetime] = None,
    ) -> Generator[StandardizedCommit, None, None]:
        if not pull_since:
            logging_helper.send_to_agent_log_file(
                f'When pulling commits for Branches, the "pull_since" argument is required for Github',
                level=logging.ERROR,
            )
            return

        try:
            login = standardized_repo.organization.login
            repo_name = self.repo_id_to_name_lookup[standardized_repo.id]
            if standardized_repo.default_branch_name:
                for j, api_commit in enumerate(
                    self.client.get_commits(
                        login=login,
                        repo_name=repo_name,
                        branch_name=standardized_repo.default_branch_name,
                        since=pull_since,
                        until=pull_until,
                    ),
                    start=1,
                ):
                    yield _standardize_commit(
                        api_commit,
                        standardized_repo,
                        standardized_repo.default_branch_name,
                        self.config.git_strip_text_content,
                        self.config.git_redact_names_and_urls,
                    )
                    if limit and j >= limit:
                        return

        except Exception as e:
            logging_helper.send_to_agent_log_file(traceback.format_exc(), level=logging.ERROR)
            logger.warning(
                f':WARN: Got exception for branch {standardized_repo.default_branch_name}: {e}. Skipping...'
            )

    def get_pr_metadata(
        self,
        standardized_repo: StandardizedRepository,
        limit: Optional[int] = None,
        pr_pull_from_date: Optional[datetime] = None,
    ) -> Generator[StandardizedPullRequestMetadata, None, None]:
        try:
            login = standardized_repo.organization.login
            repo_id = standardized_repo.id
            repo_name = self.repo_id_to_name_lookup[repo_id]

            # IF we have repo metadata for this repo and
            # IF the repo meta data indicates the latest PR for this repo
            # is older than the pr_pull_from_date, than that means we don't
            # need to pull any newly updated PRs
            repo_metadata = self.repo_id_to_repo_meta_data[repo_id]
            if repo_metadata.latest_pr_updated_at:
                if pr_pull_from_date and pr_pull_from_date >= repo_metadata.latest_pr_updated_at:
                    logging_helper.send_to_agent_log_file(
                        f'Using Repo meta data for {repo_name} (ID: {repo_id}) we were able to determine that we do not need to pull any PR data! '
                        f'pr_pull_from_data: {pr_pull_from_date}, repo_metadata.latest_pr_updated_at: {repo_metadata.latest_pr_updated_at}'
                    )
                    return
            else:
                logging_helper.send_to_agent_log_file(
                    f'Using Repo meta data for {repo_name} (ID: {repo_id}) we were able to determine that this repo has no PRs in it. No need to hit the API'
                )
                return
            for i, api_pr_metadata in enumerate(
                self.client.get_prs_metadata(login=login, repo_name=repo_name), start=1
            ):
                yield StandardizedPullRequestMetadata(
                    id=api_pr_metadata['pr']['number'],
                    updated_at=gql_format_to_datetime(api_pr_metadata['pr']['updatedAt']),
                    api_index=api_pr_metadata['cursor'],
                )
                if limit and i >= limit:
                    return

        except Exception:
            # if something happens when pulling PRs for a repo, just keep going.
            logger.warning(
                f'Problem fetching PR metadata from repo {standardized_repo.name} ({standardized_repo.id}). Skipping...'
            )
            logging_helper.send_to_agent_log_file(traceback.format_exc(), level=logging.ERROR)

    def git_provider_pr_endpoint_supports_date_filtering(self):
        return False

    def get_prs(
        self,
        standardized_repo: StandardizedRepository,
        pull_files_for_pr: bool = False,
        hash_files_for_prs: bool = False,
        limit: Optional[int] = None,
        start_cursor: Optional[Any] = None,
        start_window: Optional[datetime] = None,
        end_window: Optional[datetime] = None,  # Not used in Github
    ) -> Generator[StandardizedPullRequest, None, None]:
        try:
            login = standardized_repo.organization.login
            repo_id = standardized_repo.id
            repo_name = self.repo_id_to_name_lookup[repo_id]
            repo_metadata = self.repo_id_to_repo_meta_data[repo_id]
            if not repo_metadata.latest_pr_updated_at:
                logging_helper.send_to_agent_log_file(
                    f'Not querying for PR data, because the repo meta for {standardized_repo.name} indicates that there are no PRs'
                )
                return
            if start_window and repo_metadata.latest_pr_updated_at <= start_window:
                logging_helper.send_to_agent_log_file(
                    f'Not querying for PR data, because the repo meta for {standardized_repo.name} indicates that the latest PR '
                    f'was updated on {repo_metadata.latest_pr_updated_at} before our pull from window {start_window}'
                )
                return

            try:
                labels_for_repository = self.client.get_labels_for_repository(
                    org_login=standardized_repo.organization.login, repo_name=repo_name
                )
                label_node_id_to_id = {
                    label['node_id']: label['id'] for label in labels_for_repository
                }
            except Exception as e:
                logging_helper.send_to_agent_log_file(
                    msg=f'Error when attempting to pull Labels for Repo {repo_name}. Error: {e}',
                    level=logging.ERROR,
                )
                logging_helper.send_to_agent_log_file(
                    msg=traceback.format_exc(), level=logging.ERROR
                )
                label_node_id_to_id = {}

            api_prs = self.client.get_prs(
                login=login,
                repo_name=repo_name,
                include_top_level_comments=self.config.jf_options.get(
                    'get_all_issue_comments', False
                ),
                start_cursor=start_cursor,
                pull_files_for_pr=pull_files_for_pr,
                hash_files_for_prs=hash_files_for_prs,
                repository_label_node_ids_to_id=label_node_id_to_id,
            )
            for j, api_pr in enumerate(
                api_prs,
                start=1,
            ):
                try:
                    yield _standardize_pr(
                        api_pr,
                        standardized_repo,
                        self.config.git_strip_text_content,
                        self.config.git_redact_names_and_urls,
                    )
                    if limit and j >= limit:
                        return
                except Exception:
                    # if something goes wrong with normalizing one of the prs - don't stop pulling. try
                    # the next one.
                    pr_id = f' {api_pr["id"]}' if api_pr else ''
                    logger.warning(
                        f'normalizing PR {pr_id} from repo {standardized_repo.name} ({standardized_repo.id}). Skipping...'
                    )
                    logging_helper.send_to_agent_log_file(
                        traceback.format_exc(), level=logging.ERROR
                    )

        except Exception:
            # if something happens when pulling PRs for a repo, just keep going.
            logger.warning(
                f'normalizing PRs from repo {standardized_repo.name} ({standardized_repo.id}). Skipping...'
            )
            logging_helper.send_to_agent_log_file(traceback.format_exc(), level=logging.ERROR)


'''

    Massage Functions

'''


def _standardize_user(api_user) -> StandardizedUser:
    id = api_user.get('id')
    name = api_user.get('name')
    login = api_user.get('login')
    email = api_user.get('email')
    # raw user, just have email (e.g. from a commit)
    if not id:
        return StandardizedUser(
            id=email,
            login=email,
            name=name,
            email=email,
        )

    # API user, where github matched to a known account
    return StandardizedUser(id=id, login=login, name=name, email=email)


def _standardize_team(api_team: Dict) -> StandardizedTeam:
    return StandardizedTeam(
        id=str(api_team.get('id', '')),
        name=api_team.get('name', ''),
        slug=api_team.get('slug', ''),
        description=api_team.get('description'),
        members=[_standardize_user(member) for member in api_team.get('members', []) if member],
    )


def _standardize_organization(
    api_org: Dict, redact_names_and_urls: bool
) -> StandardizedOrganization:
    return StandardizedOrganization(
        id=api_org['id'],
        login=api_org['login'],
        name=(
            api_org.get('name')
            if not redact_names_and_urls
            else GitAdapter.organization_redactor.redact_name(api_org.get('name'))
        ),
        url=api_org['url'] if not redact_names_and_urls else None,
    )


def _standardize_branch(
    api_branch, repo_id: str, is_default_branch: bool, redact_names_and_urls: bool
) -> Optional[StandardizedBranch]:
    if not api_branch:
        return None
    if not api_branch['name']:
        return None
    return StandardizedBranch(
        repo_id=repo_id,
        name=(
            api_branch['name']
            if not redact_names_and_urls
            else GitAdapter.branch_redactor.redact_name(api_branch['name'])
        ),
        sha=api_branch['target']['sha'],
        is_default=is_default_branch,
    )


def _standardize_repo(
    api_repo,
    standardized_organization: StandardizedOrganization,
    redact_names_and_urls: bool,
) -> StandardizedRepository:
    repo_name = (
        api_repo['name']
        if not redact_names_and_urls
        else GitAdapter.repo_redactor.redact_name(api_repo['name'])
    )
    url = api_repo['url'] if not redact_names_and_urls else None

    # NOTE: If a repo is completely empty, than there will be no default branch.
    # in that case, the standardized_default_branch object will be None
    standardized_default_branch = _standardize_branch(
        api_repo['defaultBranch'],
        repo_id=api_repo['id'],
        redact_names_and_urls=redact_names_and_urls,
        is_default_branch=True,
    )
    default_branch_name = standardized_default_branch.name if standardized_default_branch else None
    default_branch_sha = standardized_default_branch.sha if standardized_default_branch else None

    return StandardizedRepository(
        id=str(api_repo['id']),
        name=repo_name,
        full_name=f'{standardized_organization.login}/{repo_name}',
        url=url,
        default_branch_name=default_branch_name,
        default_branch_sha=default_branch_sha,
        is_fork=api_repo['isFork'],
        organization=standardized_organization,
    )


def _standardize_short_form_repo(
    api_repo: Dict, redact_names_and_urls: bool
) -> StandardizedShortRepository:
    repo_name = (
        api_repo['name']
        if not redact_names_and_urls
        else GitAdapter.repo_redactor.redact_name(api_repo['name'])
    )
    url = api_repo['url'] if not redact_names_and_urls else None

    return StandardizedShortRepository(id=str(api_repo['id']), name=repo_name, url=url)


def _standardize_commit(
    api_commit: Dict,
    standardized_repo: StandardizedRepository,
    branch_name: str,
    strip_text_content: bool,
    redact_names_and_urls: bool,
):
    author = _standardize_user(api_commit['author'])
    commit_url = api_commit['url'] if not redact_names_and_urls else None
    return StandardizedCommit(
        hash=api_commit['sha'],
        author=author,
        url=commit_url,
        commit_date=gql_format_to_datetime(api_commit['committedDate']),
        author_date=gql_format_to_datetime(api_commit['authoredDate']),
        message=GitAdapter.sanitize_text(api_commit['message'], strip_text_content),
        is_merge=api_commit['parents']['totalCount'] > 1,
        repo=standardized_repo.short(),  # use short form of repo
        branch_name=(
            branch_name
            if not redact_names_and_urls
            else GitAdapter.branch_redactor.redact_name(branch_name)
        ),
    )


def _get_standardized_pr_comments(
    api_comments: List[Dict], strip_text_content
) -> List[StandardizedPullRequestComment]:
    return [
        StandardizedPullRequestComment(
            user=_standardize_user(api_comment['author']),
            body=GitAdapter.sanitize_text(api_comment['body'], strip_text_content),
            created_at=parser.parse(api_comment['createdAt']),
        )
        for api_comment in api_comments
    ]


def _get_standardized_reviews(api_reviews: List[Dict]):
    return [
        StandardizedPullRequestReview(
            user=_standardize_user(api_review['author']),
            foreign_id=api_review['id'],
            review_state=PullRequestReviewState[api_review['state']].name,
        )
        for api_review in api_reviews
    ]


def _standardize_pr(
    api_pr: Dict,
    standardized_repo: StandardizedRepository,
    strip_text_content: bool,
    redact_names_and_urls: bool,
):
    base_branch_name = api_pr['baseRefName']
    head_branch_name = api_pr['headRefName']
    standardized_merge_commit = (
        _standardize_commit(
            api_pr['mergeCommit'],
            standardized_repo=standardized_repo,
            branch_name=base_branch_name,
            strip_text_content=strip_text_content,
            redact_names_and_urls=redact_names_and_urls,
        )
        if api_pr['mergeCommit']
        else None
    )

    return StandardizedPullRequest(
        id=api_pr['id'],
        additions=api_pr['additions'],
        deletions=api_pr['deletions'],
        changed_files=api_pr['changedFiles'],
        created_at=gql_format_to_datetime(api_pr['createdAt']),
        updated_at=gql_format_to_datetime(api_pr['updatedAt']),
        merge_date=gql_format_to_datetime(api_pr['mergedAt']) if api_pr['mergedAt'] else None,
        closed_date=gql_format_to_datetime(api_pr['closedAt']) if api_pr['closedAt'] else None,
        is_closed=api_pr['state'].lower() == 'closed',
        is_merged=api_pr['merged'],
        # redacted fields
        url=api_pr['url'] if not redact_names_and_urls else None,
        base_branch=(
            base_branch_name
            if not redact_names_and_urls
            else GitAdapter.branch_redactor.redact_name(base_branch_name)
        ),
        head_branch=(
            head_branch_name
            if not redact_names_and_urls
            else GitAdapter.branch_redactor.redact_name(head_branch_name)
        ),
        # sanitized fields
        title=GitAdapter.sanitize_text(api_pr['title'], strip_text_content),
        body=GitAdapter.sanitize_text(api_pr['body'], strip_text_content),
        # standardized fields
        commits=[
            _standardize_commit(
                api_commit=commit,
                standardized_repo=standardized_repo,
                branch_name=base_branch_name,
                strip_text_content=strip_text_content,
                redact_names_and_urls=redact_names_and_urls,
            )
            for commit in api_pr['commits']
        ],
        merge_commit=standardized_merge_commit,
        author=_standardize_user(api_user=api_pr['author']),
        merged_by=_standardize_user(api_user=api_pr['mergedBy']) if api_pr['mergedBy'] else None,
        approvals=_get_standardized_reviews(api_pr['reviews']),
        comments=_get_standardized_pr_comments(api_pr['comments'], strip_text_content),
        base_repo=_standardize_short_form_repo(api_pr['baseRepository'], redact_names_and_urls),
        head_repo=_standardize_short_form_repo(api_pr['baseRepository'], redact_names_and_urls),
        labels=[
            StandardizedLabel(
                id=label['id'],
                name=label['name'],
                default=label['default'],
                description=label['description'],
            )
            for label in api_pr.get('labels', [])
        ],
        files={
            file_name: StandardizedFileData(
                status=file_data['status'],
                changes=file_data['changes'],
                additions=file_data['additions'],
                deletions=file_data['deletions'],
            )
            for file_name, file_data in api_pr.get('files', {}).items()
        },
    )
